@@ -1,12 +1,13 @@
 from datetime import datetime, date, time
 from decimal import Decimal
 from operator import itemgetter
+import six
 
 from django import forms
 from django.contrib import admin, messages
 from django.contrib.admin import widgets
 from django.contrib.admin.options import csrf_protect_m
-from django.conf.urls import patterns, url
+from django.core.exceptions import PermissionDenied
 from django.forms import fields
 from django.http import HttpResponseRedirect
 from django.shortcuts import render_to_response
@@ -14,27 +15,40 @@ from django.template.context import RequestContext
 from django.utils.formats import localize
 from django.utils.translation import ugettext as _
 
+try:
+    from django.conf.urls import patterns, url
+except ImportError:  # Django < 1.4
+    from django.conf.urls.defaults import patterns, url
+
+
 from constance import config, settings
 
 
 NUMERIC_WIDGET = forms.TextInput(attrs={'size': 10})
 
 INTEGER_LIKE = (fields.IntegerField, {'widget': NUMERIC_WIDGET})
-STRING_LIKE = (fields.CharField, {'widget': forms.Textarea(attrs={'rows': 3})})
+STRING_LIKE = (fields.CharField, {
+    'widget': forms.Textarea(attrs={'rows': 3}),
+#    'required': False,
+})
 
 FIELDS = {
     bool: (fields.BooleanField, {'required': False}),
     int: INTEGER_LIKE,
-    long: INTEGER_LIKE,
     Decimal: (fields.DecimalField, {'widget': NUMERIC_WIDGET}),
     str: STRING_LIKE,
-    unicode: STRING_LIKE,
     datetime: (fields.DateTimeField, {'widget': widgets.AdminSplitDateTime}),
     date: (fields.DateField, {'widget': widgets.AdminDateWidget}),
     time: (fields.TimeField, {'widget': widgets.AdminTimeWidget}),
     float: (fields.FloatField, {'widget': NUMERIC_WIDGET}),
 }
 
+if not six.PY3:
+    FIELDS.update({
+        long: INTEGER_LIKE,
+        unicode: STRING_LIKE,
+    })
+
 
 class ConstanceForm(forms.Form):
     def __init__(self, *args, **kwargs):
@@ -58,17 +72,17 @@ class ConstanceAdmin(admin.ModelAdmin):
         return patterns('',
             url(r'^$',
                 self.admin_site.admin_view(self.changelist_view),
-                name='%s_%s_changelist' % info
-            ),
+                name='%s_%s_changelist' % info),
             url(r'^$',
                 self.admin_site.admin_view(self.changelist_view),
-                name='%s_%s_add' % info
-            ),
+                name='%s_%s_add' % info),
         )
 
     @csrf_protect_m
     def changelist_view(self, request, extra_context=None):
         # First load a mapping between config name and default value
+        if not self.has_change_permission(request, None):
+            raise PermissionDenied
         default_initial = ((name, opts['default'])
                            for name, opts in settings.CONFIG.iteritems())
         # Then update the mapping with actually values from the backend
@@ -108,15 +122,12 @@ class ConstanceAdmin(admin.ModelAdmin):
                 'help_text': _(help_text),
                 'value': localize(value),
                 'modified': value != default,
-                'form_field': form[name]
+                'form_field': form[name],
             })
         context['config'].sort(key=itemgetter('name'))
-        context_instance = RequestContext(request, current_app=self.admin_site.name)
-
-        if settings.HAS_GRAPPELLI:
-            template_name = 'admin/constance/change_list_grappelli.html'
-        else:
-            template_name = 'admin/constance/change_list.html'
+        context_instance = RequestContext(request,
+                                          current_app=self.admin_site.name)
+        template_name = 'admin/constance/change_list.html'
         return render_to_response(template_name, context,
                                   context_instance=context_instance)
 
@@ -129,18 +140,22 @@ class ConstanceAdmin(admin.ModelAdmin):
     def has_change_permission(self, request, obj=None):
         if settings.SUPERUSER_ONLY:
             return request.user.is_superuser
-        else:
-            return super(ConstanceAdmin, self).has_change_permission(request, obj)
+        return super(ConstanceAdmin, self).has_change_permission(request, obj)
 
 
 class Config(object):
     class Meta(object):
         app_label = 'constance'
-        module_name = 'config'
+        object_name = 'Config'
+        model_name = module_name = 'config'
         verbose_name_plural = 'config'
         get_ordered_objects = lambda x: False
         abstract = False
         swapped = False
+
+        def get_change_permission(self):
+            return 'change_%s' % self.model_name
+
     _meta = Meta()
 
 
